/* ************************************
   |docname| - Style sheet for docutils
   ************************************
   This style sheet provides a nicer look than Docutils' default style sheet. It's not intended for use with Sphinx, whose styles are much better.

   This file was copied from `PyInstaller <https://github.com/pyinstaller/pyinstaller/blob/a21f0660eddff0c33b3a24fea8ddeba13da9bf61/doc/stylesheets/default.css>`_.
 */

body {
	font-family: Verdana, Arial, sans-serif;
	font-size: 10pt;
	background-color: white;
}

tt, span.literal {
	font-family: monospace;
	font-size: 0.95em;
	background-color: #ECF0F3;
	padding: 0 1px;
}

div.document {
	padding-left: 1em;
	padding-right: 1em;
	/** padding: 0 20px 30px;  this overrides the prior 2 lines */
}
dl, table {
	margin-left: 2em;
	margin-right: 2em;
}

h1, h2, h3, h4, h5, h6 {
    background-color: #F2F2F2;
    border-bottom: 1px solid #CCCCCC;
    color: #20435C !important;
    font-family: 'Trebuchet MS',sans-serif;
    font-weight: normal;
    margin: 20px -20px 10px;
    padding: 3px 0 3px 10px;
}

h1 {
	text-align: center;
	font-variant: small-caps;
}
div#contents p.topic-title.first {
	font-size: large;
	font-family: 'Trebuchet MS',sans-serif;
	font-weight: normal;
}
div.contents {
	float:left;
	margin-left: -1em;
	width: 35%;
	background: white;
	margin-right: 2em;
	padding-right: 0.5em;
	border: 1px solid black;
	}
a:link, a:visited { color: #006699; }
a:hover { background-color:#006699; color:white; }

h1 a, h2 a, h3 a, h4 a, h5 a, h6 a,
h1 a:hover, h2 a:hover, h3 a:hover, h4 a:hover, h5 a:hover, h6 a:hover,
h1 a:visited, h2 a:visited, h3 a:visited, h4 a:visited, h5 a:visited, h6 a:visited {
	text-decoration: inherit;
	background-color: inherit;
	color:inherit;
}
h2 a:hover::after, h3 a:hover::after {
	color: red;
	content: " back to contents";
	text-decoration: unterline;
	font-size: x-small;
	margin-left: 1em;
}

table.docutils {
    border: 0 solid #DDCCEE;
    border-collapse: collapse;
}
table.docutils td, table.docutils th {
    background-color: #EEEEFF;
    border-left: 0 none;
    padding: 2px 5px;
}
table.docutils td p.last, table.docutils th p.last {
    margin-bottom: 0;
}
table.field-list td, table.field-list th {
    border: 0 none !important;
}
table.footnote td, table.footnote th {
    border: 0 none !important;
}
table.docutils th {
    background-color: #EEDDEE;
    border-top: 1px solid #CCAACC;
}

table.docinfo {
    border: 0 solid black;
    border-collapse: collapse;
}
table.docinfo th {
    background-color: #EEDDEE;
    border: solid thin #CCAACC;
    border-top: 0;
    border-left: 0;
    padding: 2px 5px;
}
table.docinfo td {
    background-color: #EEEEFF;
    border: solid thin #EEDDFF;
    border-width: 0 0 thin;
    padding: 2px 5px;
}


table.option-list td.option-group {
   padding-left: 0;
   background-color: #EEDDEE;
}
table.option-list td.option-group[colspan="2"] {
    background-color: #EEEEFF;
}
table.option-list td + td {
    padding-bottom: 0.5em;
}
table.option-list kbd {
   padding: 3px;
   background-color: #EEDDEE;
   white-space: nowrap;
}

th {
	padding-right: 5px;
	text-align: left;
  	background: #FFD596;
	font-weight: bold;
}
th.head {
    text-align: center;
}

td {
	background: white;
	text-align: left;
}


dl {
    margin-bottom: 15px;
}

dt {
	white-space: nowrap;
	font-weight: bold;
}

pre {
  background-color: #EEFFCC;
  border-color: #AACC99;
  border-style: solid none;
  border-width: 1px medium;
  color: #333333;
  line-height: 120%;
  padding: 0.5em;
  margin-left: 4em;
  margin-right: 4em;
  font-family: monospace;
  font-size: 0.95em;
}

/*--- sidebar --- */

div.sidebar {
	margin-left: 1em;
	margin-bottom: 1em;
	font-size: smaller;
	border: medium outset;
	padding: 0.5em 1em;  /* needed by IE */
	background-color: #ffffee;
	width: 35%;
	float: right;
	clear: right;
}

p.sidebar-title {
	font-weight: bold;
	font-size: larger;
	color: #006699;
	font-variant: small-caps;
}

div.note {
    border: solid thin black;
    background-color: #ffffee;
    width: 90%;
    padding: 0 0.5em;
}

div.note p.admonition-title {
    font-weight: bold;
    float: left;
    margin-right: 1em;
}

div.note p.admonition-title:after {
    content: ": "
}

/*
 * END OF FILE
 */
