/* .. Copyright (C) 2012-2020 Bryan A. Jones.

    This file is part of CodeChat.

    CodeChat is free software: you can redistribute it and/or modify it under
    the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    CodeChat is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along
    with CodeChat.  If not, see <http://www.gnu.org/licenses/>.

   ******************************************************************
   |docname| - Style sheet for CodeChat with the `ReadTheDocs theme`_
   ******************************************************************
   This style sheet fixes the size of CodeChat's indent so that code and comments line up when using the `ReadTheDocs theme <https://sphinx-rtd-theme.readthedocs.io/en/stable/>`_. This is only needed for the `ReadTheDocs theme`_. To use it, add the following to your project's ``conf.py``:

   .. code-block::
    :linenos:

    html_css_files = ['CodeChat_sphinx_rtd_theme.css']
 */
div.CodeChat-indent {
    font-size: 10pt;
}

